import base64
import json
import pickle
from enum import Enum
from typing import Tuple, Sequence

__all__ = ["build_package", "read_package"]

from sorcery import assigned_names

from jord.qlive_utilities.procedures import QliveRPCMethodEnum, QliveRPCMethodMap


class SerialisationMethodEnum(Enum):
    json, pickle = assigned_names()


SERIALISATION_METHOD = SerialisationMethodEnum.pickle


def build_package(method: QliveRPCMethodEnum, *args) -> bytes:
    """

    :param method:
    :param args:
    :return:
    """
    if SERIALISATION_METHOD == SerialisationMethodEnum.pickle:
        return pickle.dumps({"method": method.value, "args": args})
    elif SERIALISATION_METHOD == SERIALISATION_METHOD.json:
        return json.dump({"method": method.value, "args": args})  # TODO: ?
        # return base64.b64encode(str({"method": method.value, "args": args}).encode("ascii"))
    else:
        raise NotImplemented


def read_package(package: bytes) -> Tuple[QliveRPCMethodEnum, Sequence[str]]:
    """

    :param package:
    :return:
    """
    if SERIALISATION_METHOD == SerialisationMethodEnum.json:
        str_dict = (
            base64.b64decode(package)
            .decode("ascii")
            .replace(
                # Json library convert string dictionary to real dictionary type. Double quotes is standard format for json
                "'",
                '"',
            )
        )
        res_dict = json.loads(str_dict)  # convert string dictionary to dict format
    elif SERIALISATION_METHOD == SerialisationMethodEnum.pickle:
        res_dict = pickle.loads(package)
    else:
        raise NotImplemented
    return QliveRPCMethodMap[QliveRPCMethodEnum(res_dict["method"])], res_dict["args"]


if __name__ == "__main__":
    print(read_package(build_package(method=QliveRPCMethodEnum.clear_all)))



import random
from itertools import cycle
from typing import Any, Iterable, Sized

from PyQt5.Qt import QColor
from qgis.core import (
    QgsVectorLayer,
    QgsSymbol,
    QgsRendererCategory,
    QgsCategorizedSymbolRenderer,
)
from warg import TripleNumber

__all__ = ["categorise_layer"]


def random_rgb(mix: TripleNumber = (255, 255, 255)) -> TripleNumber:
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    return (red, green, blue)


def random_rgba(mix: TripleNumber = (1, 1, 1, 1)) -> TripleNumber:
    red = random.randrange(0, mix[0])
    green = random.randrange(0, mix[1])
    blue = random.randrange(0, mix[2])
    alpha = random.randrange(0, mix[3])
    return (red, green, blue, alpha)


def random_color_generator() -> Any:
    while 1:
        yield random_rgb()


def categorise_layer(
    layer: QgsVectorLayer,
    field_name: str = "layer",
    iterable: Iterable = random_color_generator,
) -> None:
    if isinstance(iterable, Sized):
        iterable = cycle(iterable)
    color_iter = iter(iterable())

    render_categories = []
    for cat in layer.uniqueValues(layer.fields().indexFromName(field_name)):
        sym = QgsSymbol.defaultSymbol(layer.geometryType())
        sym.setColor(QColor(*(next(color_iter)), 255))
        render_categories.append(
            QgsRendererCategory(cat, symbol=sym, label=cat, render=True)
        )

    layer.setRenderer(QgsCategorizedSymbolRenderer(field_name, render_categories))
    layer.triggerRepaint()



from typing import Sequence, Any, Tuple

__all__ = ["pairs"]


def pairs(s: Sequence) -> Tuple[Any, Any]:
    """
    Iterate over a list in overlapping pairs.

    Usage:
        lst = [4, 7, 11, 2]
        pairs(lst) yields (4, 7), (7, 11), (11, 2)

    http://stackoverflow.com/questions/1257413/1257446#1257446


    :param s: An iterable/list
    :return: Yields a pair of consecutive elements (lst[k], lst[k+1]) of lst. Last call yields (lst[-2], lst[-1]).
    """
    i = iter(s)
    prev = next(i)

    for item in i:
        yield prev, item
        prev = item



from typing import Mapping

from PyQt5.QtGui import QColor
from warg import TripleNumber
import random
from itertools import cycle
from typing import Any, Iterable, Sized

from PyQt5.Qt import QColor
from qgis.core import (
    QgsVectorLayer,
    QgsSymbol,
    QgsRendererCategory,
    QgsCategorizedSymbolRenderer,
    QgsLineSymbol,
)
from warg import TripleNumber

__all__ = ["style_layer_from_mapping"]


def style_layer_from_mapping(
    layer: QgsVectorLayer,
    style_mapping_field_dict: Mapping,
    field_name: str = "layer",
    default_color: TripleNumber = (0, 0, 0),
    default_opacity: float = 1.0,
    default_width: float = 0.1,
) -> None:
    style_mapping = style_mapping_field_dict[field_name]

    render_categories = []
    for cat in layer.uniqueValues(layer.fields().indexFromName(field_name)):
        cat_color = default_color
        cat_opacity = default_opacity
        cat_width = default_width

        if cat in style_mapping.keys():
            style = style_mapping[cat]
            if "color" in style:
                cat_color = (
                    int(n) for n in style["color"]
                )  #  TODO: also support with AlphaChannel | Qt.GlobalColor | QGradient
            if "opacity" in style:
                cat_opacity = max(0.0, min(float(style["opacity"]), 1.0))
            if "width" in style:
                cat_width = max(0.0, float(style["width"]))

        symbol = QgsSymbol.defaultSymbol(layer.geometryType())
        symbol.setColor(QColor(*(cat_color), 255))
        symbol.setOpacity(cat_opacity)

        if isinstance(symbol, QgsLineSymbol):
            symbol.setWidth(cat_width)
        else:
            print(f"width ignored, symbol is of type: {type(symbol)}")

        render_categories.append(
            QgsRendererCategory(cat, symbol=symbol, label=cat, render=True)
        )

    layer.setRenderer(QgsCategorizedSymbolRenderer(field_name, render_categories))
    layer.triggerRepaint()



from typing import Sequence, List, Optional

from shapely.geometry import LineString, Point, MultiPoint


__all__ = ["unique_line_points", "nearest_neighbor_within"]


def unique_line_points(lines: Sequence[LineString]) -> List[Point]:
    """


    :param lines:
    :return: Return list of unique vertices from list of LineStrings.
    :rtype: List[Point]
    """

    vertices = []

    for line in lines:
        vertices.extend(list(line.coords))

    return [Point(p) for p in set(vertices)]


def nearest_neighbor_within(others: Sequence, point, max_distance) -> Optional[Point]:
    """Find the nearest point among others up to a maximum distance.


    :param others: a list of Points or a MultiPoint
    :param point: a Point
    :param max_distance: maximum distance to search for the nearest neighbor

    :return: A shapely Point if one is within max_distance, None otherwise
    :rtype: Optional[Point]
    """
    search_region = point.buffer(max_distance)
    interesting_points = search_region.intersection(MultiPoint(others))

    if not interesting_points:
        closest_point = None
    elif isinstance(interesting_points, Point):
        closest_point = interesting_points
    else:
        distances = [
            point.distance(ip) for ip in interesting_points if point.distance(ip) > 0
        ]
        closest_point = interesting_points[distances.index(min(distances))]

    return closest_point



from enum import Enum

from qgis.core import (
    QgsMultiBandColorRenderer,
    QgsPalettedRasterRenderer,
    QgsSingleBandColorDataRenderer,
    QgsSingleBandGrayRenderer,
    QgsSingleBandPseudoColorRenderer,
)

__all__ = ["QgisRendererEnum"]


class QgisRendererEnum(Enum):
    multi_band = QgsMultiBandColorRenderer
    paletted_raster = QgsPalettedRasterRenderer
    single_band_color = QgsSingleBandColorDataRenderer
    single_band_gray = QgsSingleBandGrayRenderer
    single_band_pseudo = QgsSingleBandPseudoColorRenderer



import json
import time
from enum import Enum

__all__ = ["QliveRPCMethodEnum", "QliveRPCMethodMap"]

from typing import Mapping, Any, Tuple, Optional

import geopandas.geodataframe
import numpy
import shapely.geometry.base
from warg import passes_kws_to, Number
from pandas import DataFrame
from shapely.geometry.base import BaseGeometry
from jord.geopandas_utilities import split_on_geom_type

APPEND_TIMESTAMP = True
SKIP_MEMORY_LAYER_CHECK_AT_CLOSE = True
PIXEL_SIZE = 1
DEFAULT_NUMBER = 0
CONTRAST_ENHANCE = True
DEFAULT_LAYER_NAME = "TemporaryLayer"
DEFAULT_LAYER_CRS = "EPSG:4326"
VERBOSE = False

__all__ = [
    "add_raster",
    "add_wkb",
    "add_wkts",
    "add_wkbs",
    "add_rasters",
    "add_wkt",
    "add_dataframe",
    "add_geom_layer",
    "clear_all",
    "remove_layers",
]


def add_raster(
    qgis_instance_handle: Any,
    raster: numpy.ndarray,
    name: str = DEFAULT_LAYER_NAME,
    centroid: Tuple[Number, Number] = None,
    extent_tuple: Tuple[Number, Number, Number, Number] = None,
    pixel_size: Tuple[Number, Number] = PIXEL_SIZE,
    crs_str: str = DEFAULT_LAYER_CRS,
    default_value=DEFAULT_NUMBER,
    field: str = None,
) -> None:
    """

    :param qgis_instance_handle:
    :param raster:
    :param name:
    :param centroid:
    :param extent_tuple:
    :param pixel_size:
    :param crs_str:
    :param default_value:
    :param field:
    :return: None
    :rtype: None
    """
    from qgis.core import (
        QgsRectangle,
        QgsCoordinateReferenceSystem,
        QgsRasterBlock,
        QgsRasterBandStats,
        QgsSingleBandGrayRenderer,
        QgsMultiBandColorRenderer,
        QgsContrastEnhancement,
        QgsRasterLayer,
        QgsProcessing,
    )
    from jord.qgis_utilities.numpy_utilities.data_type import get_qgis_type
    from qgis import processing

    x_size, y_size, *rest_size = raster.shape

    if len(rest_size) == 0:
        raster = numpy.expand_dims(raster, axis=-1)

    *_, num_bands = raster.shape

    data_type = get_qgis_type(raster.dtype).value

    extent = QgsRectangle()

    if extent_tuple:
        extent.setXMinimum(extent_tuple[0])
        extent.setYMinimum(extent_tuple[1])
        extent.setXMaximum(extent_tuple[2])
        extent.setYMaximum(extent_tuple[3])
    else:
        if centroid is None:
            centroid = (0, 0)  # (x_size, y_size)

        raster_half_size = (PIXEL_SIZE * x_size / 2, PIXEL_SIZE * y_size / 2)

        if False:
            raster_half_size = raster_half_size[1], raster_half_size[0]

        extent.setXMinimum(centroid[0] - raster_half_size[0])
        extent.setYMinimum(centroid[1] - raster_half_size[1])
        extent.setXMaximum(centroid[0] + raster_half_size[0])
        extent.setYMaximum(centroid[1] + raster_half_size[1])

    raster_output = processing.run(
        "qgis:createconstantrasterlayer",
        {
            "EXTENT": extent,
            "TARGET_CRS": QgsCoordinateReferenceSystem(crs_str),  # ("EPSG:2180")
            "PIXEL_SIZE": pixel_size,
            "NUMBER": default_value,
            "OUTPUT_TYPE": data_type.value,
            "IGNORE_NODATA": True,
            "OUTPUT_NODATA_VALUE": DEFAULT_NUMBER,
            "OUTPUT": QgsProcessing.TEMPORARY_OUTPUT,
        },
    )["OUTPUT"]

    if APPEND_TIMESTAMP:
        name += f"_{time.time()}"

    layer = QgsRasterLayer(raster_output, name, "gdal")
    provider = layer.dataProvider()

    if False:
        location = None
        extent = provider.extent()
        xres = (extent.width()) / provider.xSize()
        yres = (extent.height()) / provider.ySize()
        col = int((location.x() - extent.xMinimum()) / xres)
        row = int((location.y() - extent.yMinimum()) / yres)
    elif False:
        raster_extent = provider.extent()
        raster_width = provider.xSize()
        raster_height = provider.ySize()
        no_data_value = provider.srcNoDataValue(1)
    else:
        w = layer.width()
        h = layer.height()

    provider.setEditable(True)
    block = QgsRasterBlock(data_type, w, h)
    # block = provider.block(1, extent, w, h)

    for ith_band in range(0, num_bands):
        if False:
            for i in range(0, w):
                for j in range(0, h):
                    value = raster[i][j][ith_band]
                    if value == numpy.nan:
                        block.setIsNoData(i, j)
                        continue
                    value = int(value) * 255
                    block.setValue(i, j, value)
        elif True:
            for i in range(0, w):
                for j in range(0, h):
                    block.setValue(i, j, raster[i][j][ith_band])
        else:
            block.setData(bytearray(numpy.array(map(lambda x: int(x * 255), raster))))

        if VERBOSE:
            print("writing block on band", ith_band + 1)

        provider.writeBlock(block, band=ith_band + 1)

    provider.setEditable(False)
    provider.reload()

    if num_bands == 1:
        # this is needed for the min and max value to refresh in the layer panel
        renderer = layer.renderer()

        gray_renderer = QgsSingleBandGrayRenderer(provider, 1)

        if CONTRAST_ENHANCE:
            stats = provider.bandStatistics(1, QgsRasterBandStats.All, extent)
            min_value = stats.minimumValue
            max_value = stats.maximumValue

            my_enhancement = QgsContrastEnhancement()
            my_enhancement.setContrastEnhancementAlgorithm(
                QgsContrastEnhancement.StretchToMinimumMaximum, True
            )
            my_enhancement.setMinimumValue(min_value)
            my_enhancement.setMaximumValue(max_value)
            gray_renderer.setContrastEnhancement(my_enhancement)

        layer.setRenderer(gray_renderer)
    elif num_bands != 4:
        multi_color_renderer = QgsMultiBandColorRenderer(provider, 1, 2, 3)

        layer.setRenderer(multi_color_renderer)
        layer.setDefaultContrastEnhancement()
        layer.triggerRepaint()
        # iface.legendInterface().refreshLayerSymbology(layer)
    else:
        multi_color_renderer = QgsMultiBandColorRenderer(provider, 1, 2, 3)

        layer.setRenderer(multi_color_renderer)
        layer.setDefaultContrastEnhancement()
        layer.triggerRepaint()

    if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
        layer.setCustomProperty("skipMemoryLayersCheck", 1)

    qgis_instance_handle.qgis_project.addMapLayer(layer, False)
    qgis_instance_handle.temporary_group.insertLayer(0, layer)


@passes_kws_to(add_raster)
def add_rasters(qgis_instance_handle, rasters: Mapping, **kwargs) -> None:
    """

    :param qgis_instance_handle:
    :param rasters:
    :param kwargs:
    :return:
    """
    for layer_name, raster in rasters.items():
        add_raster(qgis_instance_handle, raster, name=layer_name, **kwargs)


def add_geom_layer(
    qgis_instance_handle: Any,
    geom: BaseGeometry,
    name: Optional[str] = None,
    crs: Optional[str] = None,
    fields: Mapping = None,
    index: bool = False,
    categorise_by_attribute: Optional[str] = None,
) -> None:
    """

    An example url is “Point?crs=epsg:4326&field=id:integer&field=name:string(20)&index=yes”

      :param fields: Field=name:type(length,precision) Defines an attribute of the layer. Multiple field parameters can be added to the data provider definition. Type is one of “integer”, “double”, “string”.
      :param index:     index=yes Specifies that the layer will be constructed with a spatial index
      :param qgis_instance_handle:
      :param geom:
      :param name:
      :param crs: Crs=definition Defines the coordinate reference system to use for the layer. Definition is any string accepted by QgsCoordinateReferenceSystem.createFromString()
      :return: None
      :rtype: None
    """

    from qgis.core import QgsVectorLayer, QgsFeature

    # uri = geom.type()
    # uri = geom.wkbType()
    # uri = geom.wktTypeStr()

    uri = json.loads(geom.asJson())["type"]

    if name is None:
        name = DEFAULT_LAYER_NAME

    if crs is None:
        crs = DEFAULT_LAYER_CRS

    layer_name = f"{name}"
    if APPEND_TIMESTAMP:
        layer_name += f"_{time.time()}"

    if uri == "GeometryCollection":
        gm_group = qgis_instance_handle.temporary_group.addGroup(layer_name)

        for g in geom.asGeometryCollection():  # TODO: Look into recursion?
            uri = json.loads(g.asJson())["type"]
            sub_type = uri

            if crs:
                uri += f"?crs={crs}"

            if fields:
                for k, v in fields.items():
                    uri += f"&field={k}:{v}"

            uri += f'&index={"yes" if index else "no"}'

            feat = QgsFeature()
            feat.setGeometry(g)

            sub_layer = QgsVectorLayer(uri, f"{layer_name}_{sub_type}", "memory")
            sub_layer.dataProvider().addFeatures([feat])

            if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
                sub_layer.setCustomProperty("skipMemoryLayersCheck", 1)

            qgis_instance_handle.qgis_project.addMapLayer(sub_layer, False)
            gm_group.insertLayer(0, sub_layer)
    else:
        if crs:
            uri += f"?crs={crs}"

        if fields:
            for k, v in fields.items():
                uri += f"&field={k}:{v}"

        uri += f'&index={"yes" if index else "no"}'

        feat = QgsFeature()
        feat.setGeometry(geom)

        layer = QgsVectorLayer(uri, layer_name, "memory")
        layer.dataProvider().addFeatures([feat])

        if SKIP_MEMORY_LAYER_CHECK_AT_CLOSE:
            layer.setCustomProperty("skipMemoryLayersCheck", 1)

        qgis_instance_handle.qgis_project.addMapLayer(layer, False)
        qgis_instance_handle.temporary_group.insertLayer(0, layer)


@passes_kws_to(add_geom_layer)
def add_wkb(qgis_instance_handle: Any, wkb: str, **kwargs) -> None:
    """

    :param qgis_instance_handle:
    :param wkb:
    :param kwargs:
    :return:
    """
    from qgis.core import QgsGeometry

    add_geom_layer(qgis_instance_handle, QgsGeometry.fromWkb(wkb), **kwargs)


@passes_kws_to(add_geom_layer)
def add_wkt(qgis_instance_handle: Any, wkt: str, **kwargs) -> None:
    """

    :param qgis_instance_handle:
    :param wkt:
    :param kwargs:
    :return:
    """
    from qgis.core import QgsGeometry

    add_geom_layer(qgis_instance_handle, QgsGeometry.fromWkt(wkt), **kwargs)


@passes_kws_to(add_wkb)
def add_wkbs(qgis_instance_handle: Any, wkbs: Mapping, **kwargs) -> None:
    """

    :param qgis_instance_handle:
    :param wkbs:
    :param kwargs:
    :return:
    """
    for layer_name, wkb in wkbs.items():
        add_wkb(qgis_instance_handle, wkb, name=layer_name, **kwargs)


@passes_kws_to(add_wkt)
def add_wkts(qgis_instance_handle: Any, wkts: Mapping, **kwargs) -> None:
    """

    :param qgis_instance_handle:
    :param wkts:
    :param kwargs:
    :return:
    """
    for layer_name, wkt in wkts.items():
        add_wkt(qgis_instance_handle, wkt, name=layer_name, **kwargs)


@passes_kws_to(add_geom_layer)
def add_dataframe(qgis_instance_handle: Any, dataframe: DataFrame, **kwargs) -> None:
    """

    :param qgis_instance_handle:
    :param dataframe:
    :param kwargs:
    :return:
    """
    if isinstance(dataframe, geopandas.geodataframe.GeoDataFrame):
        columns_to_include = ("layer",)
        geom_dict = split_on_geom_type(dataframe)
        for df in geom_dict.values():
            add_wkt(qgis_instance_handle, df.to_wkt())
    elif isinstance(dataframe, DataFrame):
        geometry_column = "geometry"
        if isinstance(
            dataframe[geometry_column][0], shapely.geometry.base.BaseGeometry
        ):
            a = dataframe[geometry_column][0]
            wkts = a.geom_type == "MultiPolygon"
            # TODO: Implement
        elif isinstance(dataframe[geometry_column][0], str):
            wkts = dataframe[geometry_column]
        else:
            raise NotImplemented

        for row in wkts:
            add_geom_layer(qgis_instance_handle, row)


def remove_layers(qgis_instance_handle: Any, *args) -> None:
    """

    :param qgis_instance_handle:
    :param args:
    :return:
    """
    qgis_instance_handle.on_clear_temporary()


def clear_all(qgis_instance_handle: Any) -> None:
    """

    :param qgis_instance_handle:
    :return:
    """
    remove_layers(qgis_instance_handle)
    print("CLEAR ALL!")


class QliveRPCMethodEnum(Enum):
    # add_layers = add_layers.__name__
    add_wkt = add_wkt.__name__
    add_wkb = add_wkb.__name__
    add_wkts = add_wkts.__name__
    add_wkbs = add_wkbs.__name__
    add_dataframe = add_dataframe.__name__
    add_raster = add_raster.__name__
    add_rasters = add_rasters.__name__
    remove_layers = remove_layers.__name__
    clear_all = clear_all.__name__


funcs = locals()  # In local scope for name
QliveRPCMethodMap = {e: funcs[e.value] for e in QliveRPCMethodEnum}



from qgis.core import QgsMapLayer
from warg import AlsoDecorator

__all__ = ["QLayerEditSession"]


class QLayerEditSession(AlsoDecorator):
    def __init__(self, map_layer: QgsMapLayer):
        self.map_layer = map_layer

    def __enter__(self):
        if self.map_layer:
            self.map_layer.startEditing()
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.map_layer:
            self.map_layer.commitChanges()



from qgis.core import *  # attach main QGIS library
from qgis.utils import *  # attach main python library


def single_vector_layer():
    # Attach libraries

    import os  # attach operating system library

    # Set the working directory
    wd = "C:/test"  # Set work directory
    os.chdir(wd)  # Change the directory

    # Set a variable for the current project instance
    Prj = QgsProject().instance()  # Object for current project

    # Save the project to this file name
    pnm = "Test.qgs"  # Project file name
    pnm = wd + "/" + pnm  # Concat. with path
    Prj.write(pnm)  # Save the project

    # Create an array [] object with the polygon vertices
    vrtcs = []
    vrtcs.append(QgsPointXY(396100, 8969000))
    vrtcs.append(QgsPointXY(396100, 8973900))
    vrtcs.append(QgsPointXY(397900, 8973900))
    vrtcs.append(QgsPointXY(397900, 8969000))

    # Create a polygon from the vertices
    ply_01 = QgsGeometry.fromPolygonXY([vrtcs])

    # Create a feature object then append the polygon into
    ftr = QgsFeature()
    ftr.setGeometry(ply_01)
    print(ftr.geometry())

    # Create a layer for the feature and add to the project
    lyr = QgsVectorLayer("Polygon?crs=epsg:29194", "Test", "memory")
    Prj.addMapLayers([lyr])

    # Make the layer editable, add the feature and save
    lyr.startEditing()
    lyr.addFeature(ftr)
    lyr.commitChanges()

    # Save as a shapefile
    Fl_ou = "Test.shp"
    Fl_ou = wd + "/" + Fl_ou

    options = QgsVectorFileWriter.SaveVectorOptions()
    options.driverName = "ESRI Shapefile"

    QgsVectorFileWriter.writeAsVectorFormatV2(
        lyr, Fl_ou, QgsCoordinateTransformContext(), options
    )


def multiple_raster_layers():
    myDir = r"C:\temp"
    layers = layers = iface.mapCanvas().layers()
    pipe = QgsRasterPipe()
    for layer in layers:
        extent = layer.extent()
    width, height = layer.width(), layer.height()
    renderer = layer.renderer()
    provider = layer.dataProvider()
    crs = layer.crs().toWkt()
    pipe.set(provider.clone())
    pipe.set(renderer.clone())
    file_writer = QgsRasterFileWriter("%s\\%s.tif" % (myDir, layer.name()))
    file_writer.writeRaster(pipe, width, height, extent, layer.crs())
    print
    "%s\\%s.tif" % (myDir, layer.name())



import statistics
from typing import Union, Tuple, List

from shapely.geometry import (
    LineString,
    LinearRing,
    MultiLineString,
    MultiPolygon,
    Polygon,
)
from shapely.geometry.base import BaseGeometry

__all__ = ["zero_buffer", "sanitise", "deflimmer", "clean_geometry"]

from jord.shapely_utilities.morphology import opening, closing


def zero_buffer(
    geom: BaseGeometry,
) -> Union[
    BaseGeometry
    # Point,
    # LineString,
    # Polygon,
    # MultiPolygon
]:
    return geom.buffer(0)


def deflimmer(geom: BaseGeometry, eps: float = 1e-7) -> BaseGeometry:
    return opening(closing(geom, eps), eps)


clean_geometry = unflimmer = deflimmer


def extract_poly_coords(geom: BaseGeometry) -> Tuple[List, List]:
    if geom.type == "Polygon":
        exterior_coords = geom.exterior.coords[:]
        interior_coords = []
        for interior in geom.interiors:
            interior_coords += interior.coords[:]
    elif geom.type == "MultiPolygon":
        exterior_coords = []
        interior_coords = []
        for part in geom:
            epc = extract_poly_coords(part)  # Recursive call
            exterior_coords += epc[0]
            interior_coords += epc[1]
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")
    return exterior_coords, interior_coords


def extract_poly_rings(geom: BaseGeometry) -> Tuple[List, List]:
    interior_rings = []
    exterior_rings = []
    if isinstance(geom, Polygon):
        exterior_rings.append(geom.exterior)
        interior_rings.extend(geom.interiors)
    elif isinstance(geom, MultiPolygon):
        for part in geom.geoms:
            exterior_rings.append(part.exterior)
            interior_rings.extend(part.interiors)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return exterior_rings, interior_rings


def segments(curve: Union[LinearRing, LineString]) -> List[LineString]:
    return list(map(LineString, zip(curve.coords[:-1], curve.coords[1:])))


def mean_std_dev_line_length(geom: BaseGeometry) -> Tuple[float, float]:
    line_lengths = []
    if isinstance(geom, LineString):
        for segment in segments(geom):
            line_lengths.append(segment.length)
    elif isinstance(geom, MultiLineString):
        for li in geom.geoms:
            for segment in segments(li):
                line_lengths.append(segment.length)
    elif isinstance(geom, (Polygon, MultiPolygon)):
        exterior_rings, interior_rings = extract_poly_rings(geom)
        for ex_ring in exterior_rings:
            for segment in segments(ex_ring):
                line_lengths.append(segment.length)
        for in_ring in interior_rings:
            for segment in segments(in_ring):
                line_lengths.append(segment.length)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return (
        statistics.mean(line_lengths),
        statistics.stdev(line_lengths) if len(line_lengths) > 1 else 0,
    )


def mean_std_dev_area(geom: BaseGeometry) -> Tuple[float, float]:
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)
    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)
    else:
        raise ValueError(f"Unhandled geometry type: {repr(geom.type)}")

    return (
        statistics.mean(poly_areas),
        statistics.stdev(poly_areas) if len(poly_areas) > 1 else 0,
    )


def prune_area(geom: BaseGeometry, eps: float = 1e-7) -> BaseGeometry:
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)
    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)


def prune_rings(geom: BaseGeometry, eps: float = 1e-7) -> BaseGeometry:
    poly_areas = []
    if isinstance(geom, Polygon):
        poly_areas.append(geom.area)
    elif isinstance(geom, MultiPolygon):
        for po in geom.geoms:
            poly_areas.append(po.area)


def sanitise(geom: BaseGeometry, *args: callable) -> BaseGeometry:
    """

    #A positive distance produces a dilation, a negative distance an erosion. A very small or zero distance may sometimes be used to “tidy” a polygon.

    """

    if not len(args):
        args = (zero_buffer, deflimmer)

    for f in args:
        geom = f(geom)

    return geom


if __name__ == "__main__":

    def aishdjauisd():
        # Import constructors for creating geometry collections
        from shapely.geometry import MultiPoint, MultiLineString

        # Import necessary geometric objects from shapely module
        from shapely.geometry import Point, LineString, Polygon

        # Create Point geometric object(s) with coordinates
        point1 = Point(2.2, 4.2)
        point2 = Point(7.2, -25.1)
        point3 = Point(9.26, -2.456)
        # point3D = Point(9.26, -2.456, 0.57)

        # Create a MultiPoint object of our points 1,2 and 3
        multi_point = MultiPoint([point1, point2, point3])

        # It is also possible to pass coordinate tuples inside
        multi_point2 = MultiPoint([(2.2, 4.2), (7.2, -25.1), (9.26, -2.456)])

        # We can also create a MultiLineString with two lines
        line1 = LineString([point1, point2])
        line2 = LineString([point2, point3])
        multi_line = MultiLineString([line1, line2])
        polygon = Polygon([point2, point1, point3])

        from shapely.geometry import GeometryCollection
        from matplotlib import pyplot
        import geopandas

        geoms = GeometryCollection([multi_point, multi_point2, multi_line, polygon])
        print(mean_std_dev_line_length(geoms))
        geoms = sanitise(geoms)
        print(mean_std_dev_line_length(geoms))

        p = geopandas.GeoSeries(geoms)
        p.plot()
        pyplot.show()

    def ahsudh():
        # Import constructors for creating geometry collections
        from shapely.geometry import MultiPoint, MultiLineString

        # Import necessary geometric objects from shapely module
        from shapely.geometry import Point, LineString, Polygon

        # Create Point geometric object(s) with coordinates
        point1 = Point(2.2, 4.2)
        point2 = Point(7.2, -25.1)
        point3 = Point(9.26, -2.456)
        # point3D = Point(9.26, -2.456, 0.57)

        # Create a MultiPoint object of our points 1,2 and 3
        multi_point = MultiPoint([point1, point2, point3])

        # It is also possible to pass coordinate tuples inside
        multi_point2 = MultiPoint([(2.2, 4.2), (7.2, -25.1), (9.26, -2.456)])

        # We can also create a MultiLineString with two lines
        line1 = LineString([point1, point2])
        line2 = LineString([point2, point3])
        multi_line = MultiLineString([line1, line2])
        polygon = Polygon([point2, point1, point3])
        polygon2 = Polygon([point3, point2, point1])
        multi_polygon = MultiPolygon([polygon, polygon2])

        print(mean_std_dev_line_length(line1))
        print(mean_std_dev_line_length(line2))
        print(mean_std_dev_line_length(multi_line))
        print(mean_std_dev_line_length(polygon))
        print(mean_std_dev_line_length(polygon2))
        print(mean_std_dev_line_length(multi_polygon))

        print(mean_std_dev_area(polygon))
        print(mean_std_dev_area(polygon2))
        print(mean_std_dev_area(multi_polygon))

        from shapely.geometry import GeometryCollection
        from matplotlib import pyplot
        import geopandas

        geoms = GeometryCollection(
            [multi_point, multi_point2, multi_line, polygon, polygon2, multi_polygon]
        )

        p = geopandas.GeoSeries(geoms)
        p.plot()
        pyplot.show()

    ahsudh()
    # aishdjauisd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "heider"
__doc__ = r"""

           Created on 1/23/23
           """

__all__ = [
    "to_lines",
    "to_single_line",
    "explode_line",
    "strip_multiline_dangles",
    "strip_line_dangles",
]

import collections
from typing import Union, List, Sequence

import shapely.ops
from shapely.geometry import LineString, MultiLineString, Point, MultiPoint, box
from shapely.geometry.base import BaseGeometry

from jord.shapely_utilities.points import (
    unique_line_points,
    nearest_neighbor_within,
)


def to_single_line(s: Union[LineString, MultiLineString]) -> LineString:
    """
    assume that lines are ordered, NOTE closes of gaps!

    :param s:
    :type s: Union[LineString, MultiLineString]
    :return:
    :rtype: LineString
    """
    if isinstance(s, MultiLineString):
        out_coords = [
            list(i.coords) for i in s.geoms
        ]  # Put the subline coordinates into a list of sublists

        return LineString(
            [i for sublist in out_coords for i in sublist]
        )  # Flatten the list of sublists and use it to make a new line

    elif isinstance(s, LineString):
        return s
    else:
        raise NotImplementedError


def to_lines(geoms: Sequence[BaseGeometry]) -> List[LineString]:
    """
    Converts Shapely geoms in to Shapely LineString

    :param geoms:
    :type geoms: Sequence[BaseGeometry]
    :return:
    :rtype: List[LineString]
    """

    lines = []
    for g in geoms:
        if isinstance(g, (LineString)):
            lines.append(g)
        elif isinstance(g, (BaseGeometry)):
            boundary = g.boundary
            if isinstance(boundary, MultiLineString):
                lines.extend(to_lines(boundary.geoms))
            else:
                lines.append(boundary)
        else:
            raise NotImplementedError(f"{g, type(g)}")

    return lines


def strip_line_dangles(
    line: LineString, dangle_length_threshold: float = 0.1, iterations: int = 3
) -> LineString:
    """

    :param line:
    :type line: LineString
    :param dangle_length_threshold:
    :type dangle_length_threshold: float
    :param iterations:
    :type iterations: int
    :return: The LineString without dangles shorter than the dangle_length_threshold
    :rtype: LineString
    """

    working_line = line
    for ith_ in range(iterations):
        working_segments = []
        segments = explode_line(working_line)
        if len(segments) > 2:
            start, *rest, end = segments

            if start.length > dangle_length_threshold:
                working_segments.append(start)

            working_segments.extend(rest)

            if end.length > dangle_length_threshold:
                working_segments.append(end)

        elif len(segments) < 2:
            segment = segments[0]

            if segment.length > dangle_length_threshold:
                working_segments.append(segment)

        else:
            s1, s2 = segments

            if s1.length > dangle_length_threshold:
                working_segments.append(s1)

            if s2.length > dangle_length_threshold:
                working_segments.append(s2)

        working_line = LineString(working_segments)

    return working_line


def line_endpoints(lines: Union[List[LineString], MultiLineString]) -> MultiPoint:
    """

    :param lines:
    :type: Union[List[LineString], MultiLineString]
    :return: Returns a MultiPoint of terminal points from list of LineStrings.
    :rtype: MultiPoint
    """

    all_points = []
    if isinstance(lines, MultiLineString):
        lines = lines.geoms

    for line in lines:
        for i in [0, -1]:  # start and end point
            all_points.append(line.coords[i])

    endpoints = set(
        [item for item, count in collections.Counter(all_points).items() if count < 2]
    )  # Remove duplicates

    return MultiPoint([Point(p) for p in endpoints])


def strip_multiline_dangles(
    multilinestring: MultiLineString,
    dangle_length_threshold: float = 0.1,
    iterations: int = 3,
) -> MultiLineString:
    """

    :param multilinestring:
    :type multilinestring: MultiLineString
    :param dangle_length_threshold:
    :type dangle_length_threshold: float
    :param iterations:
    :type iterations: int
    :return:
    :rtype: MultiLineString
    """
    working_multi = multilinestring
    for ith_ in range(iterations):
        endpoints = line_endpoints(working_multi)
        working_segments = []
        for linestring in working_multi.geoms:
            segments = explode_line(linestring)
            if len(segments) > 2:
                start, *rest, end = segments
                if start.intersects(endpoints):
                    if start.length > dangle_length_threshold:
                        working_segments.append(start)
                else:
                    working_segments.append(start)

                working_segments.extend(rest)

                if end.intersects(endpoints):
                    if end.length > dangle_length_threshold:
                        working_segments.append(end)
                else:
                    working_segments.append(end)
            elif len(segments) < 2:
                segment = segments[0]
                if segment.intersects(endpoints):
                    if segment.length > dangle_length_threshold:
                        working_segments.append(segment)
                else:
                    working_segments.append(segment)
            else:
                s1, s2 = segments
                if s1.intersects(endpoints):
                    if s1.length > dangle_length_threshold:
                        working_segments.append(s1)
                else:
                    working_segments.append(s1)

                if s2.intersects(endpoints):
                    if s2.length > dangle_length_threshold:
                        working_segments.append(s2)
                else:
                    working_segments.append(s2)

        working_multi = MultiLineString(working_segments)

    return working_multi


def explode_line(line: LineString) -> List[LineString]:
    """

    :param line:
    :return:
    """
    out = []
    for pt1, pt2 in zip(
        line.coords, line.coords[1:]
    ):  # iterate from first cord, iterate from second coords to get
        # endpoints of each segment
        out.append(LineString([pt1, pt2]))
    return out


def find_isolated_endpoints(
    lines: Sequence[Union[LineString, MultiLineString]],
) -> Sequence[Point]:
    """
    Find endpoints of lines that don't touch another line.

    :param lines: A list of LineStrings or a MultiLineString
    :return: A list of line end Points that don't touch any other line of lines
    """

    isolated_endpoints = []
    for i, line in enumerate(lines):
        other_lines = lines[:i] + lines[i + 1 :]
        for q in [0, -1]:
            endpoint = Point(line.coords[q])
            if any(endpoint.touches(another_line) for another_line in other_lines):
                continue
            else:
                isolated_endpoints.append(endpoint)
    return isolated_endpoints


def snappy_endings(
    lines: Union[LineString, MultiLineString], max_distance: float
) -> Sequence[Union[LineString, MultiLineString]]:
    """
    Snap endpoints of lines together if they are at most max_length apart.


    :param lines: A list of LineStrings or a MultiLineString
    :param max_distance: maximum distance two endpoints may be joined together
    :return:
    :rtype: Sequence[Union[LineString, MultiLineString]]
    """

    # initialize snapped lines with list of original lines
    # snapping points is a MultiPoint object of all vertices
    snapped_lines = [line for line in lines]
    snapping_points = unique_line_points(snapped_lines)

    # isolated endpoints are going to snap to the closest vertex
    isolated_endpoints = find_isolated_endpoints(snapped_lines)

    # only move isolated endpoints, one by one
    for endpoint in isolated_endpoints:
        # find all vertices within a radius of max_distance as possible
        target = nearest_neighbor_within(snapping_points, endpoint, max_distance)

        # do nothing if no target point to snap to is found
        if not target:
            continue

            # find the LineString to modify within snapped_lines and update it
        for i, snapped_line in enumerate(snapped_lines):
            if endpoint.touches(snapped_line):
                snapped_lines[i] = bend_towards(snapped_line, where=endpoint, to=target)
                break

        # also update the corresponding snapping_points
        for i, snapping_point in enumerate(snapping_points):
            if endpoint.equals(snapping_point):
                snapping_points[i] = target
                break

    # post-processing: remove any resulting lines of length 0
    snapped_lines = [s for s in snapped_lines if s.length > 0]

    return snapped_lines


def bend_towards(line: LineString, where: Point, to: Point) -> LineString:
    """
    Move the point where along a line to the point at location to.

    :param line:
    :param where: a point ON the line (not necessarily a vertex)
    :param to: a point NOT on the line where the nearest vertex will be moved to
    :return: the modified (bent) line
    """

    if not line.contains(where) and not line.touches(where):
        raise ValueError("line does not contain the point where.")

    coords = line.coords[:]
    # easy case: where is (within numeric precision) a vertex of line
    for k, vertex in enumerate(coords):
        if where.almost_equals(Point(vertex)):
            # move coordinates of the vertex to destination
            coords[k] = to.coords[0]
            return LineString(coords)

    # hard case: where lies between vertices of line, so
    # find the nearest vertex and move that one to point to
    _, min_k = min(
        (where.distance(Point(vertex)), k) for k, vertex in enumerate(coords)
    )
    coords[min_k] = to.coords[0]
    return LineString(coords)


def prune_short_lines(
    lines: Sequence[LineString], min_length: float
) -> List[LineString]:
    """
    Remove lines from a LineString shorter than min_length.

    Deletes all lines from a list of LineStrings or a MultiLineString
    that have a total length of less than min_length. Vertices of touching
    lines are contracted towards the centroid of the removed line.


    :param lines: List of LineStrings or a MultiLineString
    :param min_length: minimum length of a single LineString to be preserved
    :return:  the pruned pandas DataFrame
    """
    pruned_lines = [line for line in lines]  # converts MultiLineString to list
    to_prune = []

    for i, line in enumerate(pruned_lines):
        if line.length < min_length:
            to_prune.append(i)
            for n in intersecting_lines_idx(line, pruned_lines):
                contact_point = line.intersection(pruned_lines[n])
                pruned_lines[n] = bend_towards(
                    pruned_lines[n], where=contact_point, to=line.centroid
                )

    return [line for i, line in enumerate(pruned_lines) if i not in to_prune]


def linemerge(
    line_s: Union[LineString, MultiLineString]
) -> Union[LineString, MultiLineString]:
    """
    Merge a list of LineStrings and/or MultiLineStrings.

    Given a list of LineStrings and possibly MultiLineStrings, merge all of
    them to a single MultiLineString.

    :type line_s: LineString|MultiLineString
    :rtype:LineString|MultiLineString
    """
    lines = []
    for line in line_s.geoms:
        if isinstance(line, MultiLineString):
            # line is a multilinestring, so append its components
            lines.extend(line.geoms)
        else:
            # line is a line, so simply append it
            lines.append(line)

    return shapely.ops.linemerge(lines)


def one_linestring_per_intersection(
    lines: Sequence[LineString],
) -> Union[LineString, MultiLineString]:
    """
    Move line endpoints to intersections of line segments.

    Given a list of touching or possibly intersecting LineStrings, return a
     list of LineStrings that have their endpoints at all crossings and
    intersecting points and ONLY there.


    :param lines: A list of LineStrings or a MultiLineString
    :return: a list of LineStrings
    """
    lines_merged = shapely.ops.linemerge(lines)

    # intersecting multiline with its bounding box somehow triggers a first
    bounding_box = box(*lines_merged.bounds)

    # perform linemerge (one linestring between each crossing only)
    # if this fails, write function to perform this on a bbox-grid and then
    # merge the result
    lines_merged = lines_merged.intersection(bounding_box)
    lines_merged = shapely.ops.linemerge(lines_merged)
    return lines_merged


def intersecting_lines_idx(of: LineString, lines: Sequence[LineString]) -> List[int]:
    """Find the indices in a list of LineStrings that touch a given LineString.


    :param lines: List of LineStrings in which to search for neighbors
    :param of: the LineString, which must be touched
    :return: a list of indices, so that all lines[indices] touch the LineString of
    """
    return [k for k, line in enumerate(lines) if line.touches(of)]


def intersecting_lines(of: LineString, lines: Sequence[LineString]) -> List[LineString]:
    """
    Find the indices in a list of LineStrings that touch a given LineString.


    :param of: The LineString which must be touched
    :param lines: List of LineStrings in which to search for neighbors
    :return: list of indices, so that all lines[indices] touch the LineString of
    """
    return [line for line in (lines) if line.touches(of)]


if __name__ == "__main__":

    def iashdh():
        print(
            to_single_line(MultiLineString([[[0, 0], [0, 1]], [[0, 2], [0, 3]]]))
        )  # LINESTRING (0 0, 0 1, 0 2, 0 3)

    def ausdh():
        from shapely.geometry import MultiPolygon, Point

        pol1 = MultiPolygon([Point(0, 0).buffer(2.0), Point(1, 1).buffer(2.0)])
        pol2 = Point(7, 8).buffer(1.0)
        pols = [pol1, pol2]

        print(to_lines(pols))

    ausdh()



__all__ = []

from jord.qlive_utilities.procedures import QliveRPCMethodEnum
from jord.qlive_utilities.serialisation import build_package, read_package

if __name__ == "__main__":
    print(read_package(build_package(method=QliveRPCMethodEnum.clear_all)))
